"""
Node objects definition

There are two main types of node objects`, noth of which are subclasses
of the `_BaseNode` class:
- `Node` should be used for defining all documented taxonomic nodes.
- `DummyNode`can be used to add non-existing nodes, for example
  at missing ranks in order to normalize different branches

In addition, the `MergedNode` class serves to create references to
another `Node` for taxomic IDs that no longer exist due to being merged
with other IDs.
"""


from __future__ import annotations
from typing import Union, Optional
from .utils import _rand_id


class _BaseNode:
    """
    Base object for Node classes

    This is meant for internal use. Nodes should be instanciated from
    a subclass of `_BaseNode`, preferably `Node`.

    Parameters
    ---------
    taxid: str or int
        Taxonomic identification number
    name: str, optional
        Node name
    rank: str, optional
        Node rank
    parent: _BaseNode, optional
        The parent Node object

    Notes
    -----
    The `children` property will be dynamically populated when children Nodes
    declare a Node as parent.

    Attributes
    ----------
    taxid
    name
    rank
    children
    parent
    node_info
    """
    def __init__(self,
                 taxid: Union[str, int] = None,
                 name: Optional[str] = None,
                 rank: Optional[str] = None,
                 parent: Optional[_BaseNode] = None
                ) -> None:
        self._children = set()
        self._name = name
        self._rank = rank
        self._parent = parent
        self._taxid = str(taxid) if taxid != None else taxid

        self._updateParent()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.taxid})"

    # Property methods
    @property
    def taxid(self) -> str:
        """Taxonomic identification number"""
        return self._taxid

    @property
    def name(self) -> str:
        """Name of the taxonomic node"""
        return self._name

    @property
    def rank(self) -> str:
        """Rank of the taxonomic node"""
        return self._rank

    @property
    def parent(self) -> str:
        """Parent node"""
        return self._parent

    @property
    def children(self) -> set:
        """Children nodes"""
        return self._children

    @property
    def node_info(self) -> str:
        """
        Node information
        """
        return f"{self.__repr__()}\n" \
               f"type: {self.__class__.__name__}\n" \
               f"taxid: {self.taxid}\n" \
               f"name: {self.name}\n" \
               f"rank: {self.rank}\n" \
               f"parent: {self.parent}\n" \
               f"children: {self.children}\n"

    # Setter methods
    @taxid.setter
    def taxid(self, taxid: Union[str, int]) -> None:
        self._taxid = str(taxid)

    @name.setter
    def name(self, name: str) -> None:
        self._name = str(name)

    @rank.setter
    def rank(self, rank: str) -> None:
        self._rank = str(rank)

    @children.setter
    def children(self, children: set) -> None:
        self._children = set(children)

    @parent.setter
    def parent(self, parent: Node) -> None:
        """Set parent node and update children attribute of parent node"""
        # root node has circular reference to self.
        if parent and parent.taxid != self.taxid:
            assert isinstance(parent, _BaseNode)
            self._parent = parent
            self._updateParent()
        else:
            self._parent = None

    def isAncestorOf(self, node: Node) -> bool:
        """
        Test if the object is an ancestor of another Node.

        Parameters
        ----------
        node:
            Putative descendant node

        Examples
        --------
        >>> root = Node(1, "root", "root")
        >>> node = Node(2, "node", "rank", root)
        >>> node.isAncestorOf(root)
        False
        root.isAncestorOf(node)
        True
        """
        if not node.parent or node.parent.taxid == node.taxid:
            return False
        elif node.parent.taxid == self.taxid:
            return True
        else:
            return self.isAncestorOf(node.parent)

    def isDescendantOf(self, node: Node) -> bool:
        """
        Test if the object is an ancestor of another Node.

        Parameters
        ----------
        node:
            Putative ancestor node

        Examples
        --------
        >>> root = Node(1, "root", "root")
        >>> node = Node(2, "node", "rank", root)
        >>> node.isDescendantOf(root)
        True
        root.isDescendantOf(node)
        False
        """
        if not self.parent or self.parent.taxid == self.taxid:
            return False
        elif self.parent.taxid == node.taxid:
            return True
        else:
            return self.parent.isDescendantOf(node)

    def _updateParent(self) -> None:
        """
        Add self to parent's children list
        """
        if self._parent:
            self._parent.children.add(self)

    def _relink(self) -> None:
        """
        Bypass self by relinking children to parents
        """
        if not self.parent:
            raise TypeError("Cannot relink a root Node")

        children = self.children

        for child in children:
            child.parent = self.parent
            # Will auto update the parent node

        self.parent.children.discard(self)

    def _to_dict(self):
        """
        Create a dict of self with information to recreate the object.
        """
        dic = dict(self.__dict__)
        if self.parent:
            dic['_parent'] = dic['_parent'].taxid
        dic['type'] = self.__class__.__name__
        del dic['_children']
        return dic


class Node(_BaseNode):
    """
    Taxonomic Node

    Create a Node object contining taxonomic information
    as well as a link to parent and children nodes.

    Parameters
    ---------
    taxid: str or int
        Taxonomic identification number
    name: str, optional
        Node name
    rank: str, optional
        Node rank
    parent: _BaseNode, optional
        The parent Node object

    Notes
    -----
    The `children` property will be dynamically populated when children Nodes
    declare a Node as parent.

    Attributes
    ----------
    taxid
    name
    rank
    children
    parent
    node_info

    Examples
    --------
    >>> root = Node(1, "root", "root")
    >>> child = Node(2, "child", "child_rank", root)

    >>> child.taxid
    '2'
    >>> child.rank
    'child_rank'
    >>> child.name
    'child'

    >>> child.parent
    Node object:
            Taxid: 1
            Name: root
            Rank: root
            Parent: None

    >>> root.children
    [Node object:
            Taxid: 2
            Name: child
            Rank: child_rank
            Parent: 1]
    """

    def __init__(self,
                 taxid: Union[str, int],
                 name: Optional[str] = None,
                 rank: Optional[str] = None,
                 parent: Optional[_BaseNode] = None) -> None:
        super().__init__(taxid, name, rank, parent)

    # Property methods
    @property
    def taxid(self) -> str:
        """Taxonomic identification number"""
        return super().taxid

    @property
    def name(self) -> str:
        """Name of the taxonomic node"""
        return super().name

    @property
    def rank(self) -> str:
        """Rank of the taxonomic node"""
        return super().rank

    @property
    def parent(self) -> str:
        """Parent node"""
        return super().parent

    @property
    def children(self) -> list:
        """Children nodes"""
        return super().children

    @property
    def node_info(self) -> str:
        """
        Node information
        """
        return super().node_info

    # Setter methods
    @taxid.setter
    def taxid(self, taxid: Union[str, int]) -> None:
        super(Node, self.__class__).taxid.fset(self, taxid)

    @name.setter
    def name(self, name: str) -> None:
        super(Node, self.__class__).name.fset(self, name)

    @rank.setter
    def rank(self, rank: str) -> None:
        super(Node, self.__class__).rank.fset(self, rank)

    @children.setter
    def children(self, children: list) -> None:
        super(Node, self.__class__).children.fset(self, children)

    @parent.setter
    def parent(self, parent: Node) -> None:
        """Set parent node and update children attribute of parent node"""
        super(Node, self.__class__).parent.fset(self, parent)


class DummyNode(_BaseNode):
    """
    A placeholder for a non-existing Node.

    Will be assigned a random hash id in place of a taxid
    upon creation. Can be inserted between two existing nodes.

    Parameters
    ---------
    taxid: int or str, optional
        Taxonomic identification number
    name: str, optional
        Node name
    rank: str, optional
        Node rank
    parent: _BaseNode, optional
        The parent Node object

    Attributes
    ----------
    taxid
    name
    rank
    children
    parent
    node_info
    """
    def __init__(self,
                 taxid: Optional[Union[str, int]] = None,
                 name: Optional[str] = None,
                 rank: Optional[str] = None,
                 parent: Optional[_BaseNode] = None) -> None:
        if not taxid:
            taxid = _rand_id() # generating random taxid
        super().__init__(taxid, name, rank, parent)

    def insertNode(self, parent: _BaseNode, child: _BaseNode) -> None:
        """
        Insert the dummy node between parent and child

        Parameters
        ----------
        parent: _BaseNode
            Upstream node
        child: _BaseNode
            Downstream node

        Returns
        ------
        None
        """
        child.parent = self
        parent.children.remove(child)
        self.parent = parent

    # Property methods
    @property
    def taxid(self) -> str:
        """Taxonomic identification number"""
        return super().taxid

    @property
    def name(self) -> str:
        """Name of the taxonomic node"""
        return super().name

    @property
    def rank(self) -> str:
        """Rank of the taxonomic node"""
        return super().rank

    @property
    def parent(self) -> str:
        """Parent node"""
        return super().parent

    @property
    def children(self) -> list:
        """Children nodes"""
        return super().children

    @property
    def node_info(self) -> str:
        """Node information"""
        return super().node_info

    # Setter methods
    @taxid.setter
    def taxid(self, taxid: Union[str, int]) -> None:
        super(DummyNode, self.__class__).taxid.fset(self, taxid)

    @name.setter
    def name(self, name: str) -> None:
        super(DummyNode, self.__class__).name.fset(self, name)

    @rank.setter
    def rank(self, rank: str) -> None:
        super(DummyNode, self.__class__).rank.fset(self, rank)

    @children.setter
    def children(self, children: list) -> None:
        super(DummyNode, self.__class__).children.fset(self, children)

    @parent.setter
    def parent(self, parent: Node) -> None:
        """Set parent node and update children attribute of parent node"""
        super(DummyNode, self.__class__).parent.fset(self, parent)


class MergedNode:
    """
    Simple class linking to an instance of (a subclass of) _BaseNode

    Represents a taxonomic node that has been merged with another node and is therefore not
    part of the taxonomy anymore.
    This is not a subclass of _BaseNode.

    Parameters
    ---------
    taxid: int or str
        Taxonomic identification number
    new_node: int or str
        Taxid of node this node has been merged with
    *args: optional
        ignored
    **kwargs: optional
        ignored

    Attributes
    ----------
    taxid
    new_node

    Note
    ----
    `new_node` is provided as a taxid and not as an instance of a Node class.
    An Error will be raised upon trying to access a MergedNode from Taxonomy object if it is linked to a non-existing Node.
    """
    def __init__(self, taxid: Union[str, int], new_node: Union[str, int], *args, **kwargs) -> None:
        self._taxid = str(taxid)
        self._new_node = str(new_node)

    @property
    def taxid(self) -> str:
        """Taxonomic identification number"""
        return self._taxid

    @property
    def new_node(self) -> str:
        """Node this node has been merged with"""
        return self._new_node

    @taxid.setter
    def taxid(self, taxid: Union[str, int]) -> None:
        self._taxid = str(taxid)

    @new_node.setter
    def new_node(self, new_node: Union[str, int]) -> None:
        self._new_node = str(new_node)



"""
Misc. utility functions
"""


import warnings
import random
import string


def linne() -> list:
    """
    Linnean Taxonomy:

    Returns
    -------
    list
        ['species',
        'genus',
        'family',
        'order',
        'class',
        'phylum',
        'kingdom']
    """
    return ['species',
            'genus',
            'family',
            'order',
            'class',
            'phylum',
            'kingdom']


def _rand_id(ncar: int = 8) -> str:
    """Random hash"""
    return ''.join([random.choice(
                    string.ascii_letters + string.digits)
                    for n in range(ncar)])


def _deprecation(depr, replace):
    """Standard deprecation warning"""
    warnings.warn(
        f"'{depr}' is pending deprecation, use the '{replace}' instead",
        DeprecationWarning, stacklevel=2
    )



import os
import unittest
from tempfile import TemporaryDirectory


import taxidTools


current_path = os.path.dirname(__file__)
nodes = os.path.join(current_path, "data", "mininodes.dmp")
rankedlineage = os.path.join(current_path, "data", "minirankedlineage.dmp")
merged = os.path.join(current_path, "data", "minimerged.dmp")


class TestTaxdump(unittest.TestCase):

    def setUp(self):
        self.workdir = TemporaryDirectory()
        self.parent = taxidTools.Node(taxid = 0, name = "root", rank = "root", parent = None)
        self.child = taxidTools.Node(taxid = 1, name = "child", rank = "child", parent = self.parent)
        self.txd = taxidTools.Taxonomy({'0': self.parent, '1': self.child})

    def tearDown(self):
        self.workdir.cleanup()

    def test_factory_dict(self):
        self.txd = taxidTools.Taxonomy({'0': self.parent, '1': self.child})
        self.assertEqual(len(self.txd.keys()), 2)

    def test_factory_add_node(self):
        self.txd = taxidTools.Taxonomy()
        self.txd.addNode(self.child)
        self.txd.addNode(self.parent)
        self.assertEqual(len(self.txd.keys()), 2)

    def test_factory_list(self):
        self.txd = taxidTools.Taxonomy.from_list([self.parent, self.child])
        self.assertEqual(len(self.txd.keys()), 2)

    def test_factory_taxdump(self):
        self.txd = taxidTools.read_taxdump(nodes, rankedlineage, merged)
        self.assertEqual(self.txd["9913"].parent.taxid, "9903")

        ancestry = taxidTools.Lineage(self.txd["9903"])
        self.assertEqual(len(ancestry), 29)
        self.assertEqual(ancestry[-1].taxid, "1")

        self.assertEqual(self.txd["999999"], self.txd["9103"])

    def test_IO_json(self):
        self.txd = taxidTools.read_taxdump(nodes, rankedlineage, merged)
        self.txd.write(os.path.join(self.workdir.name, "test.json"))
        self.reload = taxidTools.read_json(os.path.join(self.workdir.name, "test.json"))

        self.assertEqual(self.txd["999999"], self.txd["9103"])

        ancestry = taxidTools.Lineage(self.reload["9903"])
        self.assertEqual(len(ancestry), 29)
        self.assertEqual(ancestry[-1].taxid, "1")

        self.txd.filterRanks(['genus', 'none'])
        self.txd.write(os.path.join(self.workdir.name, "test2.json"))
        test2 = taxidTools.read_json(os.path.join(self.workdir.name, "test2.json"))
        ancestry = taxidTools.Lineage(test2["9903"])
        self.assertIsInstance(ancestry[1], taxidTools.DummyNode)

    def test_getters(self):
        self.assertEqual(self.txd.getName(1), "child")
        self.assertEqual(self.txd.getRank(1), "child")
        self.assertEqual(self.txd.getParent(1).taxid, "0")

    def test_getAncestry(self):
        lin = self.txd.getAncestry(1)
        self.assertEqual(len(lin), 2)
        self.assertEqual(lin[0].taxid, "1")
        self.assertEqual(lin[1].taxid, "0")

    def test_ancestry_tests(self):
        self.assertTrue(self.txd.isAncestorOf(0,1))
        self.assertFalse(self.txd.isAncestorOf(1,0))
        self.assertFalse(self.txd.isAncestorOf(1,1))

        self.assertTrue(self.txd.isDescendantOf(1,0))
        self.assertFalse(self.txd.isDescendantOf(0,1))
        self.assertFalse(self.txd.isDescendantOf(1,1))

    def test_copy(self):
        self.new = self.txd.copy()
        self.txd.data = {}
        self.assertIsNone(self.txd.get('0'))
        self.assertIsNotNone(self.new.get('0', None))

    def test_InvalidNodeError(self):
        # Making sure InvalidNodeError can also be caught as a KeyError
        self.assertRaises(taxidTools.InvalidNodeError, self.txd.__getitem__, "notataxid")
        self.assertRaises(KeyError, self.txd.__getitem__, "notataxid")
    
    def test_MergedNnode(self):
        self.merged = taxidTools.MergedNode(10, 1)
        self.txd = taxidTools.Taxonomy.from_list([self.parent, self.child, self.merged])
        self.assertEqual(self.txd['1'], self.txd['10'])
        # assign a non-existing node and raise anerror
        self.merged = taxidTools.MergedNode(11, 99)
        self.assertRaises(taxidTools.InvalidNodeError, self.txd.__getitem__, "11")



import unittest
import taxidTools

class TestNode(unittest.TestCase):

    def setUp(self):
        self.node = taxidTools.Node(taxid = 123456)
        self.midnode = taxidTools.Node(taxid = 2, parent = self.node)
        self.lownode = taxidTools.Node(taxid = 3, parent = self.midnode)

    def test_taxid(self):
        self.assertIsInstance(self.node.taxid, str)
        self.assertEqual(self.node.taxid, "123456")

    def test_name(self):
        name = "TestName"
        self.node.name = name
        self.assertEqual(self.node.name, name)

    def test_rank(self):
        rank = "TestRank"
        self.node.rank = rank
        self.assertEqual(self.node.rank, rank)

    def test_parent(self):
        parent1 = taxidTools.Node(taxid = 789)
        self.node.parent = parent1
        self.assertEqual(self.node.parent.taxid, "789")

    def test_children(self):
        self.assertEqual(self.node.children, {self.midnode})

    def test_ancestry(self):
        self.assertEqual(self.lownode.isDescendantOf(self.node), True)
        self.assertEqual(self.node.isDescendantOf(self.lownode), False)
        self.assertEqual(self.lownode.isAncestorOf(self.node), False)
        self.assertEqual(self.node.isAncestorOf(self.lownode), True)

    def test_dummy_insert(self):
        dummy = taxidTools.DummyNode()
        dummy.insertNode(parent = self.midnode, child = self.lownode)
        self.assertEqual(dummy.parent, self.midnode)
        self.assertEqual(dummy.children, {self.lownode})
        self.assertEqual(self.midnode.children, {dummy})
        self.assertEqual(self.lownode.parent, dummy)

    def test_relink(self):
        self.midnode._relink()
        self.assertEqual(self.lownode.parent, self.node)
        self.assertEqual(len(self.node.children), 1)
        self.assertIn(self.lownode, self.node.children)

    def test_merged(self):
        self.merged = taxidTools.MergedNode(1234, 2)
        self.assertEqual(self.merged.new_node, "2")
        self.assertEqual(self.merged.taxid, "1234")


import unittest
import taxidTools
from taxidTools.Taxonomy import _insert_nodes_recc, _insert_dummies


class TestComplexTree(unittest.TestCase):
    # Test Tree
    #
    # 0(-/-  /- 001) for testing filterRanks
    # |- 1
    # |  |- 11
    # |  |- 12
    # |      |- 121
    # |      |- 122
    # |- 2
    #    |- 21
    #    |- 22
    #    |- 23
    # (|-3) for testing _insert_nodes_recc

    def setUp(self):
        self.node0 = taxidTools.Node(taxid = 0, name = "root", rank = "root", parent = None)
        self.node1 = taxidTools.Node(taxid = 1, name = "node1", rank = "rank1", parent = self.node0)
        self.node2 = taxidTools.Node(taxid = 2, name = "node2", rank = "rank1", parent = self.node0)
        self.node11 = taxidTools.Node(taxid = 11, name = "node11", rank = "rank2", parent = self.node1)
        self.node12 = taxidTools.Node(taxid = 12, name = "node12", rank = "rank2", parent = self.node1)
        self.node21 = taxidTools.Node(taxid = 21, name = "node21", rank = "rank2", parent = self.node2)
        self.node22 = taxidTools.Node(taxid = 22, name = "node22", rank = "rank2", parent = self.node2)
        self.node23 = taxidTools.Node(taxid = 23, name = "node23", rank = "rank2", parent = self.node2)
        self.node121 = taxidTools.Node(taxid = 121, name = "node121", rank = "rank3", parent = self.node12)
        self.node122 = taxidTools.Node(taxid = 122, name = "node122", rank = "rank3", parent = self.node12)

        nodes = {
            "0" : self.node0,
            "1" : self.node1,
            "2" : self.node2,
            "11" : self.node11,
            "12" : self.node12,
            "21" : self.node21,
            "22" : self.node22,
            "23" : self.node23,
            "121" : self.node121,
            "122" : self.node122
            }

        self.txd = taxidTools.Taxonomy(nodes)

    def test_insertdummy(self):
        dummy = taxidTools.DummyNode(name= 'dummy', rank = 'rank2.5')
        dummy.insertNode(self.node2, self.node21)

        self.assertEqual(self.node21.parent, dummy)
        self.assertEqual(dummy.parent, self.node2)
        self.assertEqual(dummy.children, {self.node21})
        self.assertCountEqual(self.node2.children, {dummy, self.node22, self.node23})
        self.assertEqual(taxidTools.Lineage(self.node21), [self.node21, dummy, self.node2, self.node0])
        self.assertCountEqual(taxidTools.Lineage(self.node22), [self.node22, self.node2, self.node0])

    def test_relink(self):
        self.node2._relink()
        self.assertEqual(self.node21.parent, self.node0)
        self.assertEqual(self.node22.parent, self.node0)
        self.assertEqual(self.node23.parent, self.node0)
        self.assertCountEqual(self.node0.children, {self.node1, self.node21, self.node22, self.node23})
        self.assertCountEqual(taxidTools.Lineage(self.node22), [self.node22,self.node0])
        self.assertCountEqual(taxidTools.Lineage(self.node21), [self.node21,self.node0])
        self.assertCountEqual(taxidTools.Lineage(self.node23), [self.node23,self.node0])

    def test_consens(self):
        self.assertEqual(self.txd.consensus(["11", "12", "21", "22", "23"], 1).taxid,
                         "0")
        self.assertEqual(self.txd.consensus(["11", "12", "21", "22", "23"], 0.6).taxid,
                         "2")
        self.assertEqual(self.txd.consensus(["11", "12", "21", "22"], 0.51).taxid,
                         "0")
        self.assertEqual(self.txd.consensus(["11", "11", "12", "22"], 0.75).taxid,
                         "1")
        self.assertEqual(self.txd.consensus(["11", "11", "11", "22", "12"], 0.51).taxid,
                         "11")
        self.assertEqual(self.txd.consensus(["121", "121", "122", "22", "12"], 0.51).taxid,
                         "12")
        self.assertEqual(self.txd.consensus(["121", "121", "23", "22", "22"], 0.51).taxid,
                         "2")
        self.assertEqual(self.txd.lca(["11", "11", "12", "22"]).taxid,
                         "0")
        self.assertEqual(self.txd.lca(["11", "11", "12"]).taxid,
                         "1")
        with self.assertRaises(taxidTools.InvalidNodeError):
            self.txd.consensus(["121", "121", "23", "22", "22", "notataxid"], 0.51)
        self.assertEqual(
            self.txd.consensus(["121", "121", "23", "22", "22", "notataxid"], 0.51, ignore_missing=True).taxid,
            "2")


    def test_consensus_dummynodes(self):
        node0 = taxidTools.Node(0)
        node1 = taxidTools.Node(1, parent = node0)
        dummy1 = taxidTools.DummyNode(parent = node0)
        node2 = taxidTools.Node(2, parent = node1)
        node3 = taxidTools.Node(3, parent = dummy1)
        node4 = taxidTools.Node(4, parent = dummy1)
        tax = taxidTools.Taxonomy.from_list([node0, node1, dummy1, node2, node3, node4])
        cons = tax.consensus(["2", "3", "4"], 0.51)
        self.assertEqual(cons, node0)


    def test_dist(self):
        self.assertEqual(self.txd.distance("11", "12"), 2)
        self.assertEqual(self.txd.distance("11", "21"), 4)
        self.assertEqual(self.txd.distance("11", "2"), 3)
        self.assertEqual(self.txd.distance("11", "1"), 1)
        self.assertEqual(self.txd.distance("121", "22"), 5)

    def test_listDescendant(self):
        self.assertSetEqual(set(self.txd.listDescendant(1)),
                            set([self.node11, self.node12, self.node121, self.node122]))
        self.assertEqual(self.txd.listDescendant(11), set())
        self.assertEqual(set(self.txd.listDescendant(1, ranks=['rank3'])),
                        set([self.node121, self.node122]))

    def test_prune(self):
        self.txd.prune(1)
        ids = [node.taxid for node in self.txd.values()]
        self.assertSetEqual(set(ids), {"0", "1", "11", "12", "121", "122"})
        self.assertEqual(self.node0.children, {self.node1})

        self.new = self.txd.prune(121, inplace=False)
        ids = [node.taxid for node in self.new.values()]
        self.assertSetEqual(set(ids), {"0", "1", "12", "121"})
        ids = [node.taxid for node in self.txd.values()]
        self.assertSetEqual(set(ids), {"0", "1", "11", "12", "121", "122"})

        self.txd.prune(11)
        ids = [node.taxid for node in self.txd.values()]
        self.assertSetEqual(set(ids), {"11", "1", "0"})

    def test_filter(self):
        node001 = taxidTools.Node('001', name = "node001", rank = "rank3", parent = self.node0)
        self.txd.addNode(node001)
        self.txd.filterRanks(ranks=['rank3', 'rank1'])
        self.assertEqual(len(self.txd), 8)
        # test relinking
        self.assertEqual(self.node121.parent, self.node1)
        self.assertIsInstance(node001.parent, taxidTools.DummyNode)

        self.new = self.txd.filterRanks(ranks=["rank1"], inplace=False)
        self.assertEqual(len(self.txd), 8)
        self.assertEqual(len(self.new), 4)

    def test_insert_dummies(self):
        new = _insert_dummies(self.node1, 'newrank')
        self.assertEqual(len(new), 2)
        for node in self.node1.children:
            self.assertIsInstance(node, taxidTools.DummyNode)

    def test_insert_nodes_recc(self):
        node3 = taxidTools.Node('3', name = "node3", rank = "rank1", parent = self.node0)
        self.txd.addNode(node3)
        new_nodes = _insert_nodes_recc(self.node0, ['rank3', 'rank2', 'rank1.5', 'rank1'])
        self.assertEqual(len(new_nodes), 12)
        with self.assertRaises(ValueError):
            _insert_nodes_recc(self.node0, ['root'])



"""
Factory functions for instanciating `taxidTools.Taxonomy` objects
"""

import json
from typing import Iterator, Optional
from .Taxonomy import Taxonomy
from .Node import Node, DummyNode, _BaseNode, MergedNode


def read_taxdump(
    nodes: str, rankedlineage: str,
    merged: Optional[str] = None,
    skip_synth: Optional[bool] = False,
    skip_env: Optional[bool] = False,
    skip_nodiv: Optional[bool] = True,
    ) -> Taxonomy:
    """
    Read a Taxonomy from the NCBI`s taxdump files

    Parameters
    ----------
    nodes: str
        Path to the nodes.dmp file
    rankedlineage: str
        Path to the rankedlineage.dmp file
    merged: str, optional
        Path to the merged.mp file
    skip_synth: bool, optional
        Skip taxa in the 'Synthetic and Chimeric' division
    skip_env: bool, optional
        Skip taxa in the 'Environmental samples' division
    skip_nodiv: bool, optional
        Skip taxa in the 'Unassigned' division

    Returns
    -------
    taxidTools.Taxonomy

    Examples
    --------
    >>> tax = read_taxdump("nodes.dmp', 'rankedlineage.dmp')
    >>> tax = read_taxdump("nodes.dmp', 'rankedlineage.dmp', merged='merged.dmp', skip_env=True, skip_synth=True)

    See Also
    --------
    read_json
    """
    txd = {}
    parent_dict = {}

    # Skipping divisions
    skip_list = []
    if skip_synth:
        skip_list += [7]
    if skip_env:
        skip_list += [11]
    if skip_nodiv:
        skip_list += [8]

    # Creating nodes
    for line in _parse_dump(nodes):
        if line[4] in skip_list:
            continue
        txd[line[0]] = Node(taxid=line[0], rank=str(line[2]))
        parent_dict[str(line[0])] = line[1]  # storing parent id

    # Add names from rankedlineage
    for line in _parse_dump(rankedlineage):
        txd[line[0]].name = line[1]

    # Update parent info
    for k, v in parent_dict.items():
        txd[k].parent = txd[v]

    # parsing merged
    if merged:
        for line in _parse_dump(merged):
            txd[line[0]] = MergedNode(line[0],line[1])

    return Taxonomy(txd)


def read_json(path: str) -> Taxonomy:
    """
    Load a Taxonomy from a previously exported json file.

    Parameters
    ----------
    path: str
        Path of file to load

    Returns
    -------
    taxidTools.Taxonomy

    See Also
    --------
    taxidTools.Taxonomy.write
    read_taxdump
    """
    # parse json
    with open(path, 'r') as fi:
        parser = json.loads(fi.read())

    txd = {}
    parent_dict = {}

    # Create nodes from records
    for record in parser:
        class_call = eval(record.pop('type'))
        if '_parent' in record:
            parent_dict[record['_taxid']] = record.pop('_parent')

        # Class attributes are hidden and therefore start with "_"
        # Init takes same named arguments with the "_"
        txd[record['_taxid']] = class_call(
            **{k[1:]: v for k, v in record.items()}
        )

    # Update parent info
    for k, v in parent_dict.items():
        if v:
            txd[k].parent = txd[v]

    return Taxonomy(txd)


def _parse_dump(filepath: str) -> Iterator:
    """
    Dump file line iterator, returns a yields of fields
    """
    with open(filepath, 'r') as dmp:
        for line in dmp:
            yield [item.strip() for item in line.split("|")]



"""
Taxonomy object definition
"""


from __future__ import annotations
from typing import Union, Iterator, Optional, Any
from collections import UserDict, Counter
from copy import copy, deepcopy
import json
from .Node import Node, DummyNode, _BaseNode, MergedNode
from .Lineage import Lineage
from .utils import linne, _deprecation
from .exceptions import InvalidNodeError


class Taxonomy(UserDict):
    """
    Stores Taxonomy nodes and their relationships

    A Taxonomy is instanciated as a dictionnary and
    each Node can be accessed by its taxid.
    A Taoxonomy object can be instanciated directly from a dictionnary,
    iteratively with the method `Taxonomy.addNode` method or from a
    list of taxdump files..

    Attributes
    ----------
    root
    data: dict
        data store

    Raises
    ------
    taxidTools.InvalidNodeError
        If trying to access a Node that doesn't exist with a bracket expression

    Notes
    -----
    Taxonomy objects are mutable and some methods will modify the
    underlying Node objects.
    Do a deep copy or use the Taxonomy.copy() method if you wish to keep the original object.

    A Taxonomy always assumes a unique root node.

    See Also
    --------
    Taxonomy.from_list: load a Taxonomy object from a list of Node
    read_taxdump: load a Taxonomy object from taxdump files
    read_json: load a Taxonomy from a previously exported json file
    Taxonomy.addNode: add a Node to a Taxonomy

    Examples
    --------
    >>> root = Node(1, "root", "root")
    >>> branch1 = Node(11, "node11", "middle", root)
    >>> branch2 = Node(12, "node12", "middel", root)
    >>> leaf1 = Node(111, "node111", "leaf", branch1)
    >>> leaf2 = Node(112, "node112", "leaf", branch1)
    >>> leaf3 = Node(121, "node121", "leaf", branch2)
    >>> leaf4 = Node(13, "node13", "leaf", root)

    From a dictionnary of Nodes:

    >>> tax = Taxonomy({"1" : root,
    ...     11: branch1,
    ...     12: branch2,
    ...     111: leaf1,
    ...     112: leaf2,
    ...     121: leaf3,
    ...     13: leaf4})

    Instanciate from a list:

    >>> tax = Taxonomy.from_list(
        [root, branch1, branch2, leaf1, leaf2, leaf3, leaf4])

    Or iteratively:

    >>> tax = Taxonomy()
    >>> for node in [root, branch1, branch2, leaf1, leaf2, leaf3, leaf4]:
    ...     tax.addNode(node)
    ...

    Or from the taxdump files:

    >>> tax = read_taxdump("nodes.dmp', 'rankedlineage.dmp', 'merged.dmp')
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # create name dict for backward lookup
        self._namedict = {}
        for k, v in self.items():
            if v.name:
                self._namedict[v.name] = k

    def __getitem__(self, key: str) -> Node:
        """
        Element getter with brackets

        Overloading default behavior to:
        - return a specific error on non-existing key
        - handle MergedNodes to return the new node
        """
        try:
            node = super().__getitem__(key)
        except KeyError:
            raise InvalidNodeError(f"There is no Node with taxid '{key}' in this Taxonomy")
        if isinstance(node, MergedNode):
            # call to getitem in case new_node was merged too
            return self.__getitem__(node.new_node)
        return node

    def __repr__(self):
        return f"{set(self.values())}"

    @classmethod
    def from_list(cls, node_list: list[_BaseNode]) -> Taxonomy:
        """
        Create a Taxonomy object from a list of Nodes

        Convert a list of Nodes into a valid Taxonomy object
        where each Node can be accessed using its taxid as key.

        Parameters
        ----------
        node_list:
            List of Node objects

        Returns
        -------
        Taxonomy

        Examples
        --------
        >>> txd = Taxonomy.from_list([Node(1), Node(2)])
        """
        for node in node_list:
            if not isinstance(node, (_BaseNode, MergedNode)):
                raise ValueError("Elements of node_list must be of type Node")

        as_dict = {node.taxid: node for node in node_list}

        return cls(as_dict)

    def copy(self) -> Taxonomy:
        """
        Create a deepcopy of the current Taxonomy instance.

        Equivalent to running copy.deepcopy()

        Returns
        -------
        Taxonomy
        """
        new = deepcopy(self.data)
        return Taxonomy(new)

    @property
    def root(self) -> Node:
        """
        Returns the root Node, assumes a single root shared by all Nodes
        """
        anynode = next(iter(self.values()))
        return Lineage(anynode)[-1]

    def addNode(self, node: Node) -> None:
        """
        Add a Node to an existing Taxonomy object.

        The Node taxid will be used a key to access element.

        Parameters
        ----------
        node: taxidTools._BaseNode
            A Node to add to the Taxonomy

        Examples
        --------
        >>> tax = Taxonomy()
        >>> tax.addNode(Node(1))
        """
        self[node.taxid] = node
        if node.name:
            self._namedict[node.name] = node.taxid

    def getTaxid(self, name: Union[int, str], value: Optional[Any] = None) -> str:
        """
        Get taxid from name


        Parameters
        ----------
        name: str or int
            Node name
        value:
            A value to return if name does not exist

        Returns
        -------
        str

        Examples
        --------
        >>> node = Node(1, "node", "rank")
        >>> tax = Taxonomy({'1':node})
        >>> tax.getTaxid('node')
        '1'
        """
        try:
            return self._namedict[name]
        except KeyError:
            return value

    def getName(self, taxid: Union[str, int], value: Optional[Any] = None) -> str:
        """
        Get taxid name

        Parameters
        ----------
        taxid: str or int
            Taxonomic identification number
        value:
            A value to return if name does not exist

        Returns
        -------
        str

        Examples
        --------
        >>> node = Node(1, "node", "rank")
        >>> tax = Taxonomy({'1':node})
        >>> tax.getName(1)
        'node'
        """
        try:
            return self[str(taxid)].name
        except InvalidNodeError:
            return value

    def getRank(self, taxid: Union[str, int], value: Optional[Any] = None) -> str:
        """
        Get taxid rank

        Parameters
        ----------
        taxid: str or int
            Taxonomic identification number
        value:
            A value to return if name does not exist

        Returns
        -------
        str

        Examples
        --------
        >>> node = Node(1, "node", "rank")
        >>> tax = Taxonomy({'1':node})
        >>> tax.getRank(1)
        'rank'
        """
        try:
            return self[str(taxid)].rank
        except InvalidNodeError:
            return value

    def getParent(self, taxid: Union[str, int], value: Optional[Any] = None) -> _BaseNode:
        """
        Retrieve parent Node

        Parameters
        ----------
        taxid: str or int
            Taxonomic identification number
        value:
            A value to return if name does not exist

        Returns
        -------
        taxidTools._BaseNode

        Examples
        --------
        >>> root = Node(1, "root", "root")
        >>> node = Node(2, "node", "rank", root)
        >>> tax = Taxonomy({'1': root, '2': node})
        >>> tax.getParent(2)
        Node(1)
        """
        try:
            return self[str(taxid)].parent
        except InvalidNodeError:
            return value

    def getChildren(self, taxid: Union[str, int], value: Optional[Any] = None) -> list[Node]:
        """
        Retrieve the children Nodes

        Parameters
        ----------
        taxid: str or int
            Taxonomic identification number
        value:
            A value to return if name does not exist

        Returns
        -------
        list

        Examples
        --------
        >>> root = Node(1, "root", "root")
        >>> node = Node(2, "node", "rank", root)
        >>> tax = Taxonomy({'1': root, '2': node})
        >>> tax.getChildren(1)
        [Node(2)]
        """
        try:
            return self[str(taxid)].children
        except InvalidNodeError:
            return value

    def getAncestry(self, taxid: Union[str, int]) -> Lineage:
        """
        Retrieve the ancestry of the given taxid

        Parameters
        ----------
        taxid: str or int
            Taxonomic identification number

        Returns
        -------
        taxidTools.Lineage

        Examples
        --------
        >>> root = Node(1, "root", "root")
        >>> node = Node(2, "node", "rank", root)
        >>> tax = Taxonomy({'1': root, '2': node})
        >>> tax.getAncestry(2)
        Lineage([Node(2), Node(1)])
        """
        return Lineage(self[str(taxid)])

    def isAncestorOf(self, taxid: Union[str, int],
                     child: Union[str, int]) -> bool:
        """
        Test if taxid is an ancestor of child

        Parameters
        ----------
        taxid: str or int
            Taxonomic identification number
        child: str or int
            Taxonomic identification number

        Returns
        -------
        bool

        See Also
        --------
        Taxonomy.isDescendantOf

        Examples
        --------
        >>> root = Node(1, "root", "root")
        >>> node = Node(2, "node", "rank", root)
        >>> tax = Taxonomy({'1': root, '2': node})
        >>> tax.isAncestorOf(1, 2)
        True
        >>> tax.isAncestorOf(2, 1)
        False
        """
        return self[str(taxid)].isAncestorOf(self[str(child)])

    def isDescendantOf(self, taxid: Union[str, int],
                       parent: Union[str, int]) -> bool:
        """
        Test if taxid is an descendant of parent

        Parameters
        ----------
        taxid: str or int
            Taxonomic identification number
        parent: str or int
            Taxonomic identification number

        Returns
        -------
        bool

        See Also
        --------
        Taxonomy.isAncestorOf

        Examples
        --------
        >>> root = Node(1, "root", "root")
        >>> node = Node(2, "node", "rank", root)
        >>> tax = Taxonomy({'1': root, '2': node})
        >>> tax.isDescendantOf(1, 2)
        False
        >>> tax.isDescendantOf(2, 1)
        True
        """
        return self[str(taxid)].isDescendantOf(self[str(parent)])

    def consensus(self, taxid_list: list[Union[str, int]],
                  min_consensus: float, ignore_missing: bool = False) -> Node:
        """
        Find a taxonomic consensus for the given
        taxid with a minimal agreement level.

        Parameters
        ----------
        taxid_list: list
            list of taxonomic identification numbers
        min_consensus: float
            minimal consensus level, between 0.5 and 1.
            Note that a minimal consensus of 1 will
            return the same result as `lastCommonNode()`
        ignore_missing: bool
            if True will ignore missing taxids form the analysis. If False (default),
            will raise an Error on missing taxids

        Returns
        -------
        taxidTools._BaseNode

        Raises
        ------
        ValueError
            If `taxid_list` contains no valid taxid and `ignore_missing` is `True`
        taxidTools.InvalidNodeError
            If `taxid_list` contains invalid taxids and `ignore_missing` is `False`

        Notes
        -----
        If no consensus can be found (for example because
        the Taxonomy contains multiple trees),
        an `IndexError` will be raised.

        See Also
        --------
        Taxonomy.lca


        Examples
        --------
        >>> node0 = Node(taxid = 0, name = "root",
                         rank = "root", parent = None)
        >>> node1 = Node(taxid = 1, name = "node1",
                         rank = "rank1", parent = node0)
        >>> node2 = Node(taxid = 2, name = "node2",
                         rank = "rank1", parent = node0)
        >>> node11 = Node(taxid = 11, name = "node11",
                          rank = "rank2", parent = node1)
        >>> node12 = Node(taxid = 12, name = "node12",
                          rank = "rank2", parent = node1)
        >>> tax = Taxonomy.from_list([node0, node1, node2, node11, node12])
        >>> tax.consensus([11, 12, 2], 0.8)
        Node(0)
        >>> tax.consensus([11, 12, 2], 0.6)
        Node(1)
        """
        # Consensus under 50% is ambiguous
        if min_consensus <= 0.5 or min_consensus > 1:
            raise ValueError(
                "Minimal consensus should be above 0.5 and under 1")

        # Filtering missing
        if ignore_missing:
            taxid_list = [self.get(str(txd), None) for txd in taxid_list]
            taxid_list = [txd.taxid for txd in taxid_list if txd]

        # Get lineages in REVERSED order
        lineages = [Lineage(self[str(txd)], ascending=False)
                    for txd in taxid_list]

        # Extend lineages so that they all are same size
        maxlen = max([len(lin) for lin in lineages])
        for lin in lineages:
            if len(lin) < maxlen:
                lin.extend([DummyNode()] * (maxlen - len(lin)))

        # Iterate over ranks descending to find last node above consensus level
        total = len(taxid_list)
        i = 0
        last = None

        while i < maxlen:
            count = Counter([lin[i] for lin in lineages])
            mostCommon = count.most_common(1)

            if mostCommon[0][1] / total >= min_consensus:
                if not(isinstance(mostCommon[0][0], DummyNode)):
                    # save current succesful consensus, and check the next one
                    last = mostCommon[0][0]
                i += 1
            else:
                break

        return last

    def lca(self, taxid_list: list[Union[str, int]], ignore_missing: bool = False) -> Node:
        """
        Get lowest common node of a bunch of taxids

        Parameters
        ----------
        taxid_list: list
            list of taxonomic identification numbers
        ignore_missing: bool
            if True will ignore missing taxids form the analysis. If False (default),
            will raise an Error on missing taxids

        Returns
        -------
        taxidTools._BasNode

        Raises
        ------
        ValueError
            If `taxid_list` contains no valid taxid and `ignore_missing` is `True`
        taxidTools.InvalidNodeError
            If `taxid_list` contains invalid taxids and `ignore_missing` is `False`

        See Also
        --------
        Taxonomy.consensus

        Examples
        --------
        >>> node0 = Node(taxid = 0, name = "root",
                         rank = "root", parent = None)
        >>> node1 = Node(taxid = 1, name = "node1",
                         rank = "rank1", parent = node0)
        >>> node2 = Node(taxid = 2, name = "node2",
                         rank = "rank1", parent = node0)
        >>> node11 = Node(taxid = 11, name = "node11",
                          rank = "rank2", parent = node1)
        >>> node12 = Node(taxid = 12, name = "node12",
                          rank = "rank2", parent = node1)
        >>> tax = Taxonomy.from_list([node0, node1, node2, node11, node12])
        >>> tax.lca([11, 12, 2])
        Node(0)
        """
        return self.consensus(taxid_list, 1, ignore_missing=ignore_missing)

    def distance(self, taxid1: Union[str, int],
                 taxid2: Union[str, int]) -> int:
        """
        Measures the distance between two nodes.

        Parameters
        ----------
        taxid1: str or int
            Taxonomic identification number
        taxid2: str or int
            Taxonomic identification number

        Returns
        -------
        int

        Examples
        --------
        >>> node0 = Node(taxid = 0, name = "root",
                         rank = "root", parent = None)
        >>> node1 = Node(taxid = 1, name = "node1",
                         rank = "rank1", parent = node0)
        >>> node2 = Node(taxid = 2, name = "node2",
                         rank = "rank1", parent = node0)
        >>> node11 = Node(taxid = 11, name = "node11",
                          rank = "rank2", parent = node1)
        >>> node12 = Node(taxid = 12, name = "node12",
                          rank = "rank2", parent = node1)
        >>> tax = Taxonomy.from_list([node0, node1, node2, node11, node12])
        >>> tax.distance(11, 2)
        3
        >>> tax.distance(11, 12)
        2
        """
        lca = self.lca([str(taxid1), str(taxid2)]).taxid

        d1 = len(Lineage(self[str(taxid1)])) - 1
        d2 = len(Lineage(self[str(taxid2)])) - 1
        dlca = len(Lineage(self[lca])) - 1

        return d1 + d2 - 2 * dlca

    def listDescendant(self, taxid: Union[str, int],
                       ranks: Optional[list] = None) -> list[Node]:
        """
        List all descendant of a node

        Parameters
        ----------
        taxid: str or int
            Taxonomic identification number
        ranks: list, optional
            list of ranks for which to return nodes

        Returns
        -------
        list

        Examples
        --------
        >>> node0 = Node(taxid = 0, name = "root",
                         rank = "root", parent = None)
        >>> node1 = Node(taxid = 1, name = "node1",
                         rank = "rank1", parent = node0)
        >>> node2 = Node(taxid = 2, name = "node2",
                         rank = "rank1", parent = node0)
        >>> node11 = Node(taxid = 11, name = "node11", #
                          rank = "rank2", parent = node1)
        >>> node12 = Node(taxid = 12, name = "node12",
                          rank = "rank2", parent = node1)
        >>> tax = Taxonomy.from_list([node0, node1, node2, node11, node12])
        >>> tax.listDescendant(1)
        [Node(11), Node(12)]
        >>> tax.listDescendant(2)
        []
        """
        current = copy(self[str(taxid)].children)
            # dont't want to update the original set!
        next = _flatten([child.children for child in current])

        all = current

        while next:
            all.update(next)
            current = next
            next = _flatten([child.children for child in current])

        if ranks:
            return [e for e in all if e.rank in ranks]
        return all

    def prune(self, taxid: Union[str, int], inplace: Optional[bool] = True) -> None:
        """
        Prune the Taxonomy at the given taxid

        Nodes not in the lineage (upwards and downwards)
        of the given taxid will be discarded.
        The Ancestors of the given taxid will be kept!

        Parameters
        ----------
        taxid: str or int
            taxid whose Lineage to keep
        inplace: bool, optional
            perfrom the operation inplace and mutate the underlying objects
            or return a mutated copy of the instance, keep the original unchanged

        Returns
        -------
        None

        Examples
        --------
        >>> node0 = Node(taxid = 0, name = "root",
                         rank = "root", parent = None)
        >>> node1 = Node(taxid = 1, name = "node1",
                         rank = "rank1", parent = node0)
        >>> node2 = Node(taxid = 2, name = "node2",
                         rank = "rank1", parent = node0)
        >>> node11 = Node(taxid = 11, name = "node11",
                          rank = "rank2", parent = node1)
        >>> node12 = Node(taxid = 12, name = "node12",
                          rank = "rank2", parent = node1)
        >>> tax = Taxonomy.from_list([node0, node1, node2, node11, node12])
        >>> tax.prune(1)

        Ancestry is kept

        >>> tax.getAncestry(11)
        Lineage([Node(11), Node(1), Node(0)])

        But other branches are gone

        >>> tax.get('2')
        None

        We can keep a copy of the:

        >>> new = tax.prune(11, inplace=False)
        >>> new.get('12')
        KeyError: '12'
        >>> tax.getAncestry('12')
        Lineage([Node(12), Node(1), Node(0)])
        """
        if inplace:
            tax = self
        else:
            tax = self.copy()

        # Getting upstream nodes
        nodes = self.getAncestry(taxid)

        # Unlinking other branches from upstream nodes
        # No need to change parents of the other nodes,
        # they will be removed from Taxonomy
        for i in range(1, len(nodes)):
            nodes[i].children = [nodes[i - 1]]

        # Adding all downstream nodes
        nodes.extend(tax.listDescendant(taxid))

        # Update taxonomy
        tax.data = {node.taxid: node for node in nodes}

        if not inplace:
            return tax

    def filterRanks(self, ranks: Optional[list[str]] = linne(), inplace: Optional[bool] = True) -> None:
        """
        Filter a Taxonomy to keep only the ranks provided as arguments.

        Modifies Taxonomy in-place to keep only the Nodes at the requested
        ranks. Nodes will be modified to conserve linkage in the Taxonomy.

        Parameters
        ----------
        ranks: list
            List of ranks to keep. Must be sorted by ascending ranks.
        inplace: bool, optional
            perfrom the operation inplace and mutate the underlying objects
            or return a mutated copy of the instance, keep the original unchanged

        Returns
        -------
        None

        Notes
        -----
        In order to enforce ankering of the Taxonomy, the root node will
        always be kept.

        Examples
        --------
        >>> node1 = Node(1, rank = "root")
        >>> node11 = Node(11, rank = "rank1", parent = node1)
        >>> node111 = Node(111, rank = "rank2", parent = node11)
        >>> node001 = Node('001', rank = "rank2", parent = node1)
        >>> tax = Taxonomy.from_list([node1, node11, node111, node001])
        >>> tax.filterRanks(['rank2', 'rank1', 'root'])
        >>> tax
        {Node(1), Node(11), DummyNode(tO841ymu), Node(111), Node(001)}

        DummyNodes are created as placeholders
        for missing ranks in the taxonomy:

        >>> node001.parent
        DummyNode(tO841ymu)

        Note that the root will be kept regardless of the input:

        >>> node1 = Node(1, rank = "root")
        >>> node11 = Node(11, rank = "rank1", parent = node1)
        >>> node111 = Node(111, rank = "rank2", parent = node11)
        >>> node001 = Node('001', rank = "rank2", parent = node1)
        >>> tax = Taxonomy.from_list([node1, node11, node111, node001])
        >>> tax.filterRanks(['rank2', 'rank1'])
        >>> tax
        {DummyNode(wmnar5QT), Node(001), Node(1), Node(11), Node(111)}

        It is also possible to keep the original instance intact and return a filtered copy:

        >>> new = tax.filterRanks(['rank1'], inplace=False)
        >>> new
        {DummyNode(wmnar5QT), Node(1), Node(11)}
        >>> tax
        {DummyNode(wmnar5QT), Node(001), Node(1), Node(11), Node(111)}
        """
        if inplace:
            tax = self
        else:
            tax = self.copy()

        # Create a list of nodes that will be used to update self
        new_nodes = []

        # First step, reduce tree
        # Remove unwanted nodes
        for node in tax.values():
            if node.rank in ranks:
                new_nodes.append(node)
            else:
                try:
                    node._relink()
                except TypeError:
                    # relinking a parent-less node raises TypeError
                    # The root will be kept whatever is asked to keep coherence
                    new_nodes.append(node)

        # Second step, expand tree
        # Reccursively add DummyNode to fill gaps
        root = tax.root
        if ranks[-1] == tax.root:
            ranks = ranks[:-1]
        new_nodes.extend(_insert_nodes_recc(root, ranks))

        # Update self
        tax.data = {node.taxid: node for node in new_nodes}

        if not inplace:
            return tax

    def write(self, path: str) -> None:
        """
        Write taxonomy to a JSON file.

        Parameters
        ----------
        path: str
            File path for the output

        See Also
        --------
        taxidTools.read_json
        """
        writer = json.dumps([node._to_dict()
                             for node in self.values()],
                            indent=4)
        with open(path, 'w') as fi:
            fi.write(writer)

    def toNewick(self, names: str = 'name') -> str:
        """
        Generate a Newock string fro the current taxonomy

        Export as Newick tree string for compatibility with other packages
        Import in ETE with format 8 (all names).
        Experimental feature

        Parameters
        ----------
        names: str
            Node attribute to use as node name, choice of 'name' or 'taxid'

        Returns
        -------
        str
        """

        def subtree(node, names):
            if names == 'name':
                namestring = str(node.name.replace(" ", "_"))
            elif names == 'taxid':
                namestring = str(node.taxid)

            if not node.children:
                return namestring
            subtrees = [subtree(child) for child in node.children]
            return f"({','.join(subtrees)}){namestring}"

        if names not in ['name', 'taxid']:
            raise ValueError("Parameter 'names' must be either 'name' or 'taxid'")

        return f"{subtree(self.root, names)};"


def _flatten(t: list) -> list:
    """
    Flatten nested list
    """
    return [item for sublist in t for item in sublist]


def _insert_nodes_recc(node: Node, ranks: list[str]) -> list[Node]:
    """
    Insert Dummy Nodes to fill gaps in ranks

    Reccursively relinks all nodes under node
    to follow the order given by ranks.
    Note that parents will be relinked with one dummy pro child!

    Notes:
    ------
    Assumes that the Taxonomy has been purged of not wanted
    ranks.

    Parameter:
    ----------
    node:
        The starting (top) node, should be the root
        when calling the function from the top level:
    ranks:
        Ascending list of ranks desired in the output.
        Should not include the root rank!

    Returns:
    --------
    list of added nodes
    """
    if 'root' in ranks:  #Move this to filter to properly check against root naming
        raise ValueError("'root' should not be included when filtering ranks. Use the Taxonomy.root property instead.")

    # if no ranks left return an empty list
    if not ranks:
        return []

    # Keep track of created dummyNodes
    new_nodes = _insert_dummies(node, ranks[-1])

    for child in node.children:
        new_nodes.extend(_insert_nodes_recc(child, ranks[:-1]))

    return new_nodes


def _insert_dummies(node, next_rank):
    dummies = []
    if node.children and next_rank:
        rerank = []
        # First check all children and keep track of
        # those that must be reranked
        for child in node.children:
            if child.rank != next_rank:
                rerank.append(child)
        # Then create dummies and insert between parent and child
        # Note: both steps uncoupled to avoid creating children
        # while iterating on children attribute.
        for child in rerank:
            dummy = DummyNode(rank=next_rank)
            dummy.insertNode(parent=node, child=child)
            dummies.append(dummy)
    elif next_rank:
        # Leaf node but still ranks left
        dummy = DummyNode(rank=next_rank, parent=node)
        dummies.append(dummy)
    return dummies



"""
Lineage object definition
"""


from __future__ import annotations
from typing import Optional
from collections import UserList
from .Node import Node, DummyNode, _BaseNode, MergedNode
from .utils import linne


class Lineage(UserList):
    """
    Taxomic Lineage

    Defines a linear and ordered succession of Nodes.
    A Lineage is created by providing a single Node that
    will be used as a base to retrieve higher Nodes.
    Ranks are ascending by default.

    Parameters
    ----------
    base_node: taxidTools._Basenode
        An instance of a `taxidTools._BaseNode`subclass from which the ancestry
        should be retrieved
    ascending: bool, optional
        Should the Lineage by sorted by ascending ranks?

    Notes
    -----
    A Lineage does not have to be continuous. Nodes can have parents that
    are not included in the Lineage, as long as Nodes in a Lineage form a
    linear path.

    Lineage methods will never modify the Node objects it contains

    Examples
    --------
    >>> root = Node(1, "root", "root")
    >>> child1 = Node(2, "child1", "child_rank", root)
    >>> child2 = Node(3, "child2", "sub_child_rank", child1)
    >>> Lineage(child2)
    Lineage([Node(3), Node(2), Node(1)])

    Lineage elements are the Node objects themselves

    >>> Lineage(child2)[-1]
    Node object:
            Taxid: 1
            Name: root
            Rank: root
            Parent: None

    A Lineage can also be descending
    >>> Lineage(child2, ascending = False)
    Lineage([Node(1), Node(2), Node(3)])
    """

    def __init__(self, base_node: _BaseNode, ascending: Optional[bool] = True) -> None:
        if not isinstance(base_node, (_BaseNode,MergedNode)):
            raise ValueError(
                "Lineage should be instanciated with a Node or list of Nodes")

        self._baseNode = base_node

        vec = [base_node]

        while vec[-1].parent:
            vec.append(vec[-1].parent)

        self.data = vec

        if not ascending:
            self.reverse()

    def filter(self, ranks: Optional[list[str]] = linne()) -> None:
        """
        Filter a Lineage to a set of specified ranks.

        Modifies a Lineage in-place.
        Lineage order will not be conserved and dummy nodes will
        be added as placeholders for missing ranks.

        Parameters
        ----------
        ranks: list, optional
            List of ranks to filter. It is assumed to be sorted
            in the same order as Lineage.

        Notes
        -----
        The Nodes are not modified by this method!
        That means that Node.parent will
        still point to the original parent Node,
        even if it was masked in the Lineage.

        Examples
        --------
        >>> root = Node(1, "root", "root")
        >>> child1 = Node(2, "child1", "child_rank", root)
        >>> child2 = Node(3, "child2", "sub_child_rank", child1)
        >>> lin = Lineage(child2)
        >>> lin.filter(["sub_child_rank", "norank", "child_rank"])
        >>> lin
        Lineage([Node(3), 'dummy', Node(2)])

        Order is not conserved!

        >>> lin = Lineage(child2)
        >>> lin.filter(["root", "sub_child_rank"])
        Lineage([Node(1), Node(3)])
        """
        nodedict = {node.rank: node for node in self if node.rank in ranks}

        new = []
        for rank in ranks:
            try:
                new.append(nodedict[rank])
            except KeyError:
                new.append(DummyNode(rank=rank))

        self.data = new

    def __repr__(self) -> str:
        return f"Lineage({[node for node in self]})"



import unittest
import taxidTools

class TestLineage(unittest.TestCase):

    def setUp(self):
        self.parent = taxidTools.Node(taxid = 0, name = "root", rank = "root", parent = None)
        self.rank1 = taxidTools.Node(taxid = 1, name = "node1", rank = "rank1", parent = self.parent)
        self.rank2 = taxidTools.Node(taxid = 2, name = "node2", rank = "rank2", parent = self.rank1)
        self.rank3 = taxidTools.Node(taxid = 3, name = "node3", rank = "rank3", parent = self.rank2)


    def test_init(self):
        # From Node
        self.lin = taxidTools.Lineage(self.rank1)
        self.linrev = taxidTools.Lineage(self.rank1, ascending = False)
        self.assertEqual(len(self.lin), 2)
        self.assertEqual(self.lin[0].taxid, "1")
        self.assertEqual(self.lin[1].taxid, "0")
        self.assertEqual(len(self.linrev), 2)
        self.assertEqual(self.linrev[0].taxid, "0")
        self.assertEqual(self.linrev[1].taxid, "1")

    def test_filter(self):
        fil1 = taxidTools.Lineage(self.rank3)
        fil1.filter(["rank3", "rank1"])
        self.assertEqual(len(fil1), 2)

        fil2 = taxidTools.Lineage(self.rank3)
        fil2.filter(["rank3", "norank", "rank1"])
        self.assertEqual(len(fil2), 3)
        self.assertTrue(isinstance(fil2[1], taxidTools.DummyNode))

        fil3 = taxidTools.Lineage(self.rank3)
        fil3.filter(["rank1", "rank3", "rank99", "rank2"])
        self.assertTrue(isinstance(fil3[2], taxidTools.DummyNode))
        self.assertEqual(fil3[0], self.rank1)
        self.assertEqual(fil3[1], self.rank3)
        self.assertEqual(fil3[3], self.rank2)



"""
Custom exception definitions
"""


class TaxonomyError(Exception):
    """Base Exception for all taxonomy related errors"""
    default_message = 'A Taxonomy related error occured'

    def __init__(self, message:str = None) -> None:
        super().__init__(message or self.default_message)


class InvalidNodeError(TaxonomyError, KeyError):
    """Raised when a Node doesn't exist"""
